import java.io.File;
import java.io.RandomAccessFile;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

class FileInfo {
	private short				YMDtag;
	private int				Year;
	private int				Mon;
	private int				Day;
	private int				Hour;
	private int				Minute;
	private int				Second;
	private int				CoodValid;
	private char				LatRef;
	private double				LatDMS;
	private char				LongRef;
	private double				LongDMS;
	private File				FilePath;
	private RandomAccessFile	JpegFile;
	private int				Orientation;		// 画像の向き
	private int				Lines;				// 画像の高さ
	private int				Width;				//	〃	幅
	private long				App1Pos;			// APP1 サイズの格納位置
	private int				App1Size;
	private long				TiffHeaderPos;
	private char				TiffEndian;			// Tiff のエンディアン、0x4949 ("II", Little) または 0x4d4d("MM", Big)
	private ReadValue			ReadV;				// Endianによって数値読み取り関数を使い分けるためのクラス
	private int				ExifLines;			// Exif Image Length
	private int				ExifWidth;			//		〃	  Width
	private String				XPTitle;			// explorerで設定されるTitle
	private List<String>		XPKeyWords;			// explorerで設定されるtag
	private String				Title;				// XMPから採取したタイトル
	private List<String>		TagList;			// XMPから採取したキーワード
	private List<String>		FoundTags;			// tagが含まれるリスト、XPKeyWord, TagList のどちらか
	
	enum SegmentType {
		OTHER,				// Other type
		APP0,
		EXIF,				// ID Code = "Exif"
		XMP,				// ID Code = "http://ns.adobe.com/xap/?.?/"
		SOS					// 0xffda Start Of Scan
	};
	
	// タグ値参照関数
	short getYMDtag() { return YMDtag; }
	int	getYear() { return Year; }
	int getMon() { return Mon; }
	int getDay() { return Day; }
	int getCoodValid() { return CoodValid; }
	char getLatRef() { return LatRef; }
	double getLatDMS() { return LatDMS; }
	char getLongRef() { return LongRef; }
	double getLongDMS() { return LongDMS; }
	int getLines() { return (ExifLines != 0) ? ExifLines : Lines; }
	int getWidth() { return (ExifWidth != 0) ? ExifWidth : Width; }
	int getOrientation() { return Orientation; }
	
	// コンストラクタ
	FileInfo(File path) {
		YMDtag = 0;
		CoodValid = 0;
		JpegFile = null;
		if (path != null) {
			FilePath = path;
			try {
				JpegFile = new RandomAccessFile(FilePath, "r");
			} catch (IOException e) {
			}
		}
	}

	// エンディアンを吸収するための抽象クラス
	private abstract class ReadValue {
		private byte Read1() {		// 1byte読み取り
			try {
				return JpegFile.readByte();
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected abstract short 	Read2();
		protected abstract int		Read4();
		protected abstract long		Read8();
		protected abstract float		Readf();
		protected abstract double	Readd();
	}
	// BigEndian用
	private class ReadValueB extends ReadValue {
		protected short Read2() {
			try {
				return JpegFile.readShort();
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected int Read4() {
			try {
				return JpegFile.readInt();
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected long Read8() {
			try {
				return JpegFile.readLong();
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected float Readf() {
			try {
				return JpegFile.readFloat();
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected double Readd() {
			try {
				return JpegFile.readDouble();
			}
			catch (IOException e) {
			}
			return 0;
		}
	}
	// LittleEndian用
	private class ReadValueL extends ReadValue {
		protected short Read2() {
			try {
				return Short.reverseBytes(JpegFile.readShort());
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected int Read4() {
			try {
				return Integer.reverseBytes(JpegFile.readInt());
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected long Read8() {
			try {
				return Long.reverseBytes(JpegFile.readLong());
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected float Readf() {
			byte[] b = new byte[4];
			int		i;
			try {
				JpegFile.read(b);
				i = ((b[3] & 0xff) << 24) | ((b[2] & 0xff) << 16) | ((b[1] & 0xff) << 8) | (b[0] & 0xff);
				return Float.intBitsToFloat(i);
			}
			catch (IOException e) {
			}
			return 0;
		}
		protected double Readd() {
			byte[] b = new byte[8];
			long   l;
			try {
				JpegFile.read(b);
				l = ((b[7] & 0xff) << 56) | ((b[6] & 0xff) << 48) | ((b[5] & 0xff) << 40) | ((b[4] & 0xff) << 32)
					| ((b[3] & 0xff) << 24) | ((b[2] & 0xff) << 16) | ((b[1] & 0xff) << 8) | (b[0] & 0xff);
				return Double.longBitsToDouble(l);
			}
			catch (IOException e) {
			}
			return 0;
		}
	}

	// 有効なJpegファイルかどうかを調べる
	boolean isValid() {
		if (JpegFile == null)
			return false;
		else
			return true;
	}
		
	private void ReadTitle(String XMPStr) {
		final String TitleStart = "<dc:title>";
		final String TitleStrStart = "<rdf:li xml:lang=\"x-default\">";
		final String TitleStrEnd = "</rdf:li>";
		final String TitleEnd = "</dc:title>";
		
		int ix = XMPStr.indexOf(TitleStart);
		if (ix < 0)
			return;
		ix += TitleStart.length();
		int endix = XMPStr.indexOf(TitleEnd, ix);
		if (endix <= ix)
			return;
		ix = XMPStr.indexOf(TitleStrStart, ix);
		if (ix <= 0 || endix <= ix)
			return;
		ix += TitleStrStart.length();
		endix = XMPStr.indexOf(TitleStrEnd, ix);
		if (endix <= ix)
			return;
		
		Title = XMPStr.substring(ix, endix);
	}
	
	// bufに読み込んだXMPデータからタグを抽出する
	private void ReadTags(String XMPStr) {
		final String KYStringStart = "<MicrosoftPhoto:LastKeywordXMP>";
		final String KYListStart = "<rdf:Bag xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\">";
		final String KYWordStart = "<rdf:li>";
		final String KYWordEnd = "</rdf:li>";
		final String KYListEnd = "</rdf:Bag>";
		// final String XYStringEnd = "</MicrosoftPhoto:LastKeywordXMP>";
		
		// タグの指定 <MicrosoftPhoto:LastKeywordXMP> を探す
		int ix = XMPStr.indexOf(KYStringStart);
		if (ix < 0) {
			return;				// タグは無い
		}
		// タグリストの開始 <rdf:Bag… を探す
		ix = XMPStr.indexOf(KYListStart, ix + KYStringStart.length());
		if (ix < 0){
			return;				// タグは無い
		}
		int endix = XMPStr.indexOf(KYListEnd, ix + KYListStart.length());	// リストの最後を探す
		if (endix < 0) {
			return;
		}
		
		ix += KYListStart.length();			// タグを探し始める位置 <rdf:li>
		TagList = new ArrayList<String>();
		while (ix < endix) {
			int endmarkix;
			ix = XMPStr.indexOf(KYWordStart, ix);
			endmarkix = XMPStr.indexOf(KYWordEnd, ix);
			if (endmarkix >= endix) {
				return;
			}
			ix += KYWordStart.length();
			TagList.add(XMPStr.substring(ix, endmarkix));
			ix = endmarkix + KYWordEnd.length();
		}
	}

	// XMPの内容を読み取る  既に "http://ns.adobe.com/xap/1.0/" を読み取っている。
	private boolean ReadXMP() {
			byte buf[] = new byte[App1Size - 31];
		
		try {
			JpegFile.read(buf, 0, 1);					// "http:..."の終端としての'\0'を読み飛ばす
			JpegFile.read(buf, 0, App1Size - 31);

			// 読み込んだXMPデータを文字列に変換
			String XMPStr = new String(buf, Charset.forName("UTF-8"));

			// タイトルを探す
			ReadTitle(XMPStr);
			
			// タグを探す
			ReadTags(XMPStr);
			
			return true;
		} catch (IOException e) {
			return false;
		}
	}
	
	// 引数の文字列の日付を整数に変換する
	private boolean getDate(byte[] buf) {
		int		y;
		int		m;
		int		d;
		int		h;
		int		mm;
		int		s;
		
		try {
			y = Integer.parseInt(new String(buf, 0, 4));
			m = Integer.parseInt(new String(buf, 5, 2));
			d = Integer.parseInt(new String(buf, 8, 2));
			h = Integer.parseInt(new String(buf, 11, 2));
			mm = Integer.parseInt(new String(buf,14, 2));
			s = Integer.parseInt(new String(buf, 17, 2));
		
		}
		catch (NumberFormatException e) {
			return false;
		}
		
		Year = y;
		Mon = m;
		Day = d;
		Hour = h;
		Minute = mm;
		Second = s;
		
		return true;
	}
	
	// GPS IFD中の3個の分数からなる緯度または経度(度分秒形式)を読み取り、小数形式にして格納する。
	private double getDMS() {
		int		numerator;
		int		denominator;
		double dms = 0;				// 
		
		numerator = ReadV.Read4();
		denominator = ReadV.Read4();
		if (denominator != 0)
			dms = (double)numerator / denominator;			// 度
		numerator = ReadV.Read4();
		denominator = ReadV.Read4();
		if (denominator != 0)
			dms += ((double)numerator / denominator) / 60;	// 分
		numerator = ReadV.Read4();
		denominator = ReadV.Read4();
		if (denominator != 0)
			dms += ((double)numerator / denominator) / 3600;	// 秒
		
		return dms;
	}
	
	// XPKeywords を読み取る
	private void getXPKeyWords(byte[] buf, int count) {
		String KeyWordString;
		final char Separator = ';';

		XPKeyWords = new ArrayList<String>();
		KeyWordString = new String(buf, Charset.forName("UTF-16LE"));		// ? ExifEndianに関わらず UTF-16LE ?
		// XPKeyWordStringをセパレータで分割して XPKeyWords に登録する
		int from = 0;
		int to = 0;
		while (from <= KeyWordString.length()) {
			to = KeyWordString.indexOf(Separator, from);
			if (to < 0) {
				// 続くキーワードはない
				XPKeyWords.add(KeyWordString.substring(from, KeyWordString.length() - 1));
				break;
			}
			else {
				XPKeyWords.add(KeyWordString.substring(from, to));
				from = to + 1;				// skip ';'
			}
		}
	}

	// XPTitleを取得
	private void getXPTitle(byte[] buf, int count) {
		// count は末尾の\u0000を含むので、バイト数を "-2" する
		XPTitle = new String(buf, 0, count - 2, Charset.forName("UTF-16LE"));			// 必ず Little Endian ?
	}
	
	// IFD 1個分のフィールドをすべて読む
	// フィールドは	tag		2
	//				type	2
	//				count	4
	//				offset	4
	// で構成される
	private boolean ReadField(int IFDCode) {
		short	tag;
		short	type;
		int		count;
		int		offset;
		long	nextpos;				// 次のフィールドの位置
		byte[]	buf = null;				// データを読み込むバッファ
		
		try {
			tag = ReadV.Read2();
			type = ReadV.Read2();
			count = ReadV.Read4();
			switch (type) {
			case 1:		// BYTE
				buf = new byte[count];
				if (count <= 4) {
					offset = (int)JpegFile.getFilePointer();
					nextpos = offset + 4;				// 4byte分(offsetの長さ)をスキップ
				}
				else {
					offset = ReadV.Read4();
					nextpos = JpegFile.getFilePointer();
					JpegFile.seek(TiffHeaderPos + offset);
				}
				JpegFile.read(buf, 0, count);
				break;
			case 2:		// ASCII
				buf = new byte[count];
				if (count > 4) {
					offset = ReadV.Read4();
					nextpos = JpegFile.getFilePointer();
					JpegFile.seek(TiffHeaderPos + offset);
				}
				else {
					offset = (int)JpegFile.getFilePointer();	// ASCII文字列の先頭位置
					nextpos = offset + 4;						// '+4'は4文字分
				}
				JpegFile.read(buf, 0, count);
				break;
			case 3:		// SHORT
				if (count > 2) {
					offset = ReadV.Read4();
					nextpos = JpegFile.getFilePointer();
				}
				else {
					offset = (int)JpegFile.getFilePointer();
					nextpos = offset + 4;
				}
				break;
			case 4:		// LONG
			case 9:		// SLONG
				if (count > 1) {
					offset = ReadV.Read4();
					nextpos = JpegFile.getFilePointer();
				}
				else {
					offset = (int)JpegFile.getFilePointer();
					nextpos = offset + 4;
				}
				break;
			case 5:		// RATIONAL
			case 10:	// SRATIONAL
				offset = ReadV.Read4();
				nextpos = JpegFile.getFilePointer();
				JpegFile.seek(TiffHeaderPos + offset);
				break;
			default:
				if (count > 4) {
					offset = ReadV.Read4();
					nextpos = JpegFile.getFilePointer();
					JpegFile.seek(TiffHeaderPos + offset);
				}
				else {
					offset = (int)JpegFile.getFilePointer();	// ASCII文字列の先頭位置
					nextpos = offset + 4;						// '+4'は4文字分
				}
				break;
			}

			//			Exif IFD		||			GPS IFD		   ||	Interoperability IFD
			if ((tag == (short)0x8769) || (tag == (short)0x8825) || (tag == (short)0xa005)) {
				// 各IFDを読み取る
				offset = ReadV.Read4();					// IFDへのオフセットを読み取る
				JpegFile.seek(TiffHeaderPos + offset);	// IFDへシーク
				ReadIFD(tag);		// 再帰する
			}
			if (IFDCode == 0) {
				// 0th IFD
				if ((tag == 0x100) && (Width == 0)) {		// Image width
					// SOFの Horizontal lines を読み取っていなければこちらを採用
					if (type == 3)
						Width = ReadV.Read2();
					else if (type == 4)
						Width = ReadV.Read4();
					else
						return false;
				}
				else if ((tag == 0x101) && (Lines == 0)) {	// Image height
					// SOFの Vertical lines を読み取っていなければこちらを採用
					if (type == 3)
						Lines = ReadV.Read2();
					else if (type == 4)
						Lines = ReadV.Read4();
					else
						return false;
				}
				else if (tag == 0x112)			// 画像の向き
					Orientation = ReadV.Read2();
				else if (tag == 0x132) {	// File change date and time
					if (getDate(buf) != true)
						return false;
					YMDtag = tag;
				}
				else if (tag == (short)0x9c9e)
					getXPKeyWords(buf, count);		// XPKeywordsを取得
				else if (tag == (short)0x9c9b)
					getXPTitle(buf, count);			// XPTitleを取得
			}
			else if (IFDCode == (short)0x8769) {
				// Exif IFDの場合
				if ((tag == (short)0x9003) || (tag == (short)0x9004)) {
					if (getDate(buf) == true)		// 日付文字列を整数に変換
						YMDtag = tag;
				}
				else if (tag == (short)0xa002) {
					if (type == 3)
						ExifWidth = ReadV.Read2();
					else
						ExifWidth = ReadV.Read4();
				}
				else if (tag == (short)0xa003) {
					if (type == 3)
						ExifLines = ReadV.Read2();
					else
						ExifLines = ReadV.Read4();
				}
						
			}
			else if (IFDCode == (short)0x8825) {
				// GPS Info IFD を読み取る
				if (tag == (short)0x1)
					LatRef = (char)buf[0];
				else if (tag == (short)0x2) {
					JpegFile.seek(TiffHeaderPos + offset);		// 緯度又は経度の格納位置へ移動
					LatDMS = getDMS();
					CoodValid++;
				}
				else if (tag == (short)0x3)
					LongRef =  (char)buf[0];
				else if (tag == (short)0x4) {
					JpegFile.seek(TiffHeaderPos + offset);		// 緯度又は経度の格納位置へ移動
					LongDMS = getDMS();
					CoodValid++;
				}
			}
			
			JpegFile.seek(nextpos);				// 次のField位置
		} catch (IOException e) {
			return false;
		}
		return true;
	}
	
	// TIFF内全てのIFDを読み取る  IFDの形式に誤りがあれば false, そうでなければ true を返す
	//	IFDCode:IFDの種別を示す。
	//		0:0th IFD
	//	 8769:Exif IFD
	//   8825:GPS IFD
	//   A005:Interoperability IFD
	private boolean ReadIFD(int IFDCode) {
		int		count;
		
		for (count = ReadV.Read2(); count > 0; count--) {
			if (ReadField(IFDCode) != true)
				return false;
		}
		return true;
	}
	

	// TIFF の情報を読み取る Endianを読み取り、TiffEndianに格納後、呼び出される
	// TIFFの形式として正しくない場合は false, そうでない場合は true を返す
	private boolean ReadTiff() {
		try {
			// Endianに従って、読み取りオブジェクトを作成する
			if (TiffEndian == 0x4d4d)
				ReadV = new ReadValueB();	// Big Endian
			else if (TiffEndian == 0x4949)
				ReadV = new ReadValueL();	// Little Endian;
			else
				return false;
			
			if (ReadV.Read2() != 0x2a)		// 0x2a 以外はエラー
				return false;
			
			long ofst = ReadV.Read4();				// TIFF Header から 0th IFD へのオフセット
			JpegFile.seek(TiffHeaderPos + ofst);	// 0th IFD へ移動
			return ReadIFD(0);						// 0th IFDを読み取る
		}
		catch (IOException e) {
			return false;
		}
	}
	
	// Application Segment 1 を読み取る
	private boolean ReadAPP1()
	{
		try {
			App1Pos = JpegFile.getFilePointer();
			App1Size = JpegFile.readChar();
			byte[] id = new byte[28];
			String idStr;

			if (JpegFile.read(id, 0, 4) < 4)			// ID文字列先頭4文字を読み取る
				return false;							// 4文字に満たない → エラー
			
			idStr = new String(id, 0, 4);			// byte列を文字列に変換
			if (idStr.equals("Exif") && (JpegFile.readChar() == 0)) {
				// Exif IDを読んだ
				TiffHeaderPos = JpegFile.getFilePointer();		// TIFFヘッダの位置を記憶
				TiffEndian = JpegFile.readChar();
				if (ReadTiff() == true) {
					JpegFile.seek(App1Pos + App1Size);
					return true;
				}
				else
					return false;
			}
			else if (idStr.equals("http")) {		// Xmp ヘッダ ? "http://ns.adobe.com/xap/1.0/"
				if (JpegFile.read(id, 4, 24) < 24)
					return false;
				idStr = new String(id, 0, 28);
				if (idStr.equals("http://ns.adobe.com/xap/1.0/") == true) {
					return ReadXMP();
				}
				else {
					// XMP でない
					return false;
				}
			}
			JpegFile.seek(App1Pos + App1Size);	// APP1の次へ移動
		} catch (IOException e) {
			return false;
		}
		
		return true;
	}
	
	// JPEG Compressed Data を読み取る SOI(Start of Image) は読み取り済み
	// 形式が誤っている場合は false, そうでない場合は true を返す
	private boolean ReadJpeg() {
		char	marker;
		
		try {
			while ((marker = JpegFile.readChar()) != 0xffe9) {
				switch (marker) {
				case 0xffe1:			// APP1
					ReadAPP1();
					break;
				case 0xffc0:			// SOF
					App1Pos = JpegFile.getFilePointer();		// 現在位置(size格納位置)を保存
					App1Size = JpegFile.readUnsignedShort();	// レングス
					JpegFile.skipBytes(1);						// サンプル精度 読み飛ばし
					Lines = JpegFile.readUnsignedShort();		// 画像の高さ
					Width = JpegFile.readUnsignedShort();		//       幅
					JpegFile.seek(App1Pos + App1Size);			// SOFの次へ移動
					break;
				case 0xffda:			// SOS	以降は SOS + Compressed Data + EOI なので、全て無視
					return true;
				case 0xffe2:			// APP2
				default:		// 上記以外は読み飛ばし
					JpegFile.seek(JpegFile.getFilePointer() + JpegFile.readChar());
					break;
				}
			}
		}
		catch (IOException e) {
			return false;
		}
		return true;
	}
	
	// true: 情報が見つかった
	// false:情報が見つからない - TIFFでもAPPnでもない
	boolean getFileInfo() {
		char	c;
		boolean retval = false;
		
		try {
			c = JpegFile.readChar();
			if ((c == 0x4949) || (c == 0x4d4d)) {
				// TIFF Header を読んだ → Uncompressed Image Data
				TiffHeaderPos = JpegFile.getFilePointer() - 2;		// TIFFヘッダの位置を記憶 すでにEndianを読み取っているので '-2' する
				TiffEndian = c;
				return ReadTiff();
			}
			else if (c == 0xffd8) {
				// SOIを読んだ → JPEG Compressed Data
				return ReadJpeg();
			}
			else
				return false;
		} catch (IOException e) {
			return retval;
		} finally {
			try {
				if (JpegFile != null)
					JpegFile.close();
			} catch (IOException e) {
			}
		}
	}
	
	// 文字列の形式で返す
	String getYMDHMSStr() {
		if (YMDtag == 0)
			return "";
		else
			return String.format("%1$04d/%2$02d/%3$02d %4$02d:%5$02d:%6$02d", Year, Mon, Day, Hour, Minute, Second);
	}
	
	String getLatDMSStr() {
		if (CoodValid == 0)
			return "";
		else
			return String.format("%1$.6f%2$c", LatDMS, LatRef);
	}
	
	String getLongDMSStr() {
		if (CoodValid == 0)
			return "";
		else
			return String.format("%1$.6f%2$c", LongDMS, LongRef);
	}
	
	String getTitle() {
		if (XPTitle != null)
			return XPTitle;
		else if (Title != null)
			return Title;
		else
			return "";
	}

	// XPKeyWords と TagList から SrchTags で指定された文字列を探す
	// 見つかれば、XPKeyWords または TagList を FoundTags に設定する
	boolean MatchCondTag(List<String> SrchTags) {
		if ((TagList != null) && (TagList.size() > 0)) {
			for(String filetag : TagList) {
				for (String srchtag : SrchTags) {
					if (filetag.indexOf(srchtag) >= 0) {
						FoundTags = TagList;
						return true;
					}
				}
			}
		}
		if ((XPKeyWords != null) && (XPKeyWords.size() > 0)) {
			for(String filetag : XPKeyWords) {
				for (String srchtag : SrchTags) {
					if (filetag.indexOf(srchtag) >= 0) {
						FoundTags = TagList;
						return true;
					}
				}
			}
		}
		
		FoundTags = null;
		return false;
	}
	// JPEGファイル中のタグを文字列にして返す
	String getTagString() {
		StringBuilder	tags = new StringBuilder();
		List<String>	list;
		int				cnt;
		int				cntTag = 0;
		int				cntKey = 0;
		
		// 検索指定された文字列を含む方を採用
		if ((FoundTags != null) && (cnt = FoundTags.size()) > 0)
			list = FoundTags;					// 検索対象のタグを発見している
		else {
			// 検索指定されていない、又は検索文字列が見つからない → 文字列が長い方を採用
			if (TagList != null)
				cntTag = TagList.size();
			if (XPKeyWords != null)
				cntKey = XPKeyWords.size();
			if ((cntTag == 0) && (cntKey == 0))
				return "";					// LastKeywordXMP も XPKeyWords も存在しない
			if (cntTag > cntKey) {
				list = TagList;				// LastKeywordXMP の方が長いので，採用する
				cnt = cntTag;
			}
			else {
				list = XPKeyWords;
				cnt = cntKey;
			}
		}
		
		for (String s: list) {
			tags.append(s);
			cnt--;
			if (cnt > 0) {
				tags.append(";");
			}
		}
		
		return tags.toString();
	}
}
